@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1985-2023, University of Amsterdam
+    Copyright (c)  1985-2024, University of Amsterdam
                               VU University Amsterdam
                               CWI, Amsterdam
                               SWI-Prolog Solutions b.v.
@@ -39,7 +39,7 @@
           [ '$find_library'/5,
             '$in_library'/3,
             '$define_predicate'/1,
-            '$update_library_index'/0,
+            '$update_library_index'/1,          % +Options
             '$autoload'/1,
 
             make_library_index/1,
@@ -136,15 +136,20 @@ user:file_search_path(autoload, Dir) :-
 :- thread_local
     silent/0.
 
-%!  '$update_library_index'
+%!  '$update_library_index'(+Options)
 %
 %   Called from make/0 to update the index   of the library for each
 %   library directory that has a writable   index.  Note that in the
 %   Windows  version  access_file/2  is  mostly   bogus.  We  assert
-%   silent/0 to suppress error messages.
+%   silent/0 to suppress error messages.  Options:
+%
+%       - system(+Boolean)
+%         Do (not) include system libraries.   Default `false`.
+%       - user(+Boolean)
+%         Do (not) include user libraries.  Default `true`.
 
-'$update_library_index' :-
-    setof(Dir, writable_indexed_directory(Dir), Dirs),
+'$update_library_index'(Options) :-
+    setof(Dir, writable_indexed_directory(Dir, Options), Dirs),
     !,
     setup_call_cleanup(
         asserta(silent, Ref),
@@ -154,7 +159,7 @@ user:file_search_path(autoload, Dir) :-
     ->  reload_library_index
     ;   true
     ).
-'$update_library_index'.
+'$update_library_index'(_).
 
 guarded_make_library_index([]).
 guarded_make_library_index([Dir|Dirs]) :-
@@ -165,11 +170,19 @@ guarded_make_library_index([Dir|Dirs]) :-
     ),
     guarded_make_library_index(Dirs).
 
-%!  writable_indexed_directory(-Dir) is nondet.
+%!  writable_indexed_directory(-Dir, +Options) is nondet.
 %
 %   True when Dir is an indexed   library  directory with a writable
 %   index, i.e., an index that can be updated.
 
+writable_indexed_directory(Dir, Options) :-
+    current_prolog_flag(home, Home),
+    writable_indexed_directory(Dir),
+    (   sub_atom(Dir, 0, _, _, Home)
+    ->  '$option'(system(true), Options, false)
+    ;   '$option'(user(true), Options, true)
+    ).
+
 writable_indexed_directory(Dir) :-
     index_file_name(IndexFile, autoload('INDEX'), [access([read,write])]),
     file_directory_name(IndexFile, Dir).
@@ -358,7 +371,7 @@ make_library_index2(Dir, Patterns) :-
     pattern_files(Patterns, DirS, Files),
     (   library_index_out_of_date(Dir, AbsIndex, Files)
     ->  do_make_library_index(AbsIndex, DirS, Files),
-        flag('$modified_index', _, true)
+        set_flag('$modified_index', true)
     ;   true
     ).
 
@@ -390,10 +403,10 @@ library_index_out_of_date(_Dir, Index, _Files) :-
 library_index_out_of_date(Dir, Index, Files) :-
     time_file(Index, IndexTime),
     (   time_file(Dir, DotTime),
-        DotTime > IndexTime
-    ;   '$member'(File, Files),
+        DotTime - IndexTime > 0.001             % compensate for jitter
+    ;   '$member'(File, Files),                 % and rounding
         time_file(File, FileTime),
-        FileTime > IndexTime
+        FileTime - IndexTime > 0.001
     ),
     !.
 
@@ -456,6 +469,11 @@ index_header(Fd):-
     format(Fd, '    Purpose: Provide index for autoload~n', []),
     format(Fd, '*/~n~n', []).
 
+%!  exports(+File, -Module, -Exports) is det.
+%
+%   Get the exports from a library as a list of PIs.
+
+:- public exports/3.                            % using by library(prolog_deps).
 exports(File, Module, Exports) :-
     (   current_prolog_flag(xref, Old)
     ->  true
@@ -463,16 +481,17 @@ exports(File, Module, Exports) :-
     ),
     setup_call_cleanup(
         set_prolog_flag(xref, true),
-        exports_(File, Module, Exports),
+        snapshot(exports_(File, Module, Exports)),
         set_prolog_flag(xref, Old)).
 
 exports_(File, Module, Exports) :-
     State = state(true, _, []),
-    (   '$source_term'(File, _,_,Term0,_,_,[syntax_errors(quiet)]),
-        (   is_list(Term0)
-        ->  '$member'(Term, Term0)
-        ;   Term = Term0
-        ),
+    (   '$source_term'(File,
+                       _Read,_RLayout,
+                       Term,_TermLayout,
+                       _Stream,
+                       [ syntax_errors(quiet)
+                       ]),
         (   Term = (:- module(M,Public)),
             is_list(Public),
             arg(1, State, true)
@@ -482,10 +501,10 @@ exports_(File, Module, Exports) :-
             fail
         ;   nb_setarg(1, State, false),
             fail
-        ;   Term = (:- export(PI)),
-            ground(PI)
-        ->  arg(3, State, E0),
-            '$append'(E0, [PI], E1),
+        ;   Term = (:- export(Export))
+        ->  phrase(export_pi(Export), PIs),
+            arg(3, State, E0),
+            '$append'(E0, PIs, E1),
             nb_setarg(3, State, E1),
             fail
         ;   Term = (:- use_foreign_library(Lib)),
@@ -497,6 +516,8 @@ exports_(File, Module, Exports) :-
         ;   Term = (:- Directive),
             nonvar(Directive)
         ->  fail
+        ;   Term == []                          % Expansion for conditionals
+        ->  fail
         ;   !
         )
     ;   true
@@ -504,6 +525,17 @@ exports_(File, Module, Exports) :-
     arg(2, State, Module),
     arg(3, State, Exports).
 
+export_pi(Var) -->
+    { var(Var) },
+    !.
+export_pi((A,B)) -->
+    !,
+    export_pi(A),
+    export_pi(B).
+export_pi(PI) -->
+    { ground(PI) },
+    [PI].
+
 
                  /*******************************
                  *            EXTENDING         *
@@ -934,10 +966,15 @@ current_autoload(M:File, Context, Term) :-
 		 *            CHECK		*
 		 *******************************/
 
+%!  warn_autoload(+TargetModule, :PI) is det.
+%
+%   @arg PI is of the shape LoadModule:Name/Arity.
+
 warn_autoload(TargetModule, PI) :-
     current_prolog_flag(warn_autoload, true),
     \+ current_prolog_flag(xref, true),
     \+ nb_current('$autoload_warning', true),
+    \+ nowarn_autoload(TargetModule, PI),
     '$pi_head'(PI, Head),
     source_file(Head, File),
     expansion_hook(P),
@@ -955,6 +992,25 @@ expansion_hook(user:goal_expansion(_,_,_,_)).
 expansion_hook(system:goal_expansion(_,_)).
 expansion_hook(system:goal_expansion(_,_,_,_)).
 
+%!  nowarn_autoload(+TargetModule, +LoadModulePI) is semidet.
+%
+%   True when LoadModule:'$nowarn_autoload'(PI,TargetModule)  is defined
+%   and true.
+%
+%   @tbd As is, these  facts  must  be   defined  by  the  library being
+%   autoloaded. Possibly we want a specific autoload declaration. As all
+%   this only affects the Prolog libraries,   we can always change this.
+%   One option might be this, where `How`   is one of `true`, `false` or
+%   `warning`.
+%
+%      :- autoloadable(PI, How)
+
+nowarn_autoload(TargetModule, LoadModule:PI) :-
+    NoWarn = LoadModule:'$nowarn_autoload'(PI,TargetModule),
+    '$c_current_predicate'(_, NoWarn),
+    \+ '$get_predicate_attribute'(NoWarn, imported, _From),
+    call(NoWarn).
+
 
                  /*******************************
                  *             REQUIRE          *
