@@ -350,7 +350,7 @@ path_sep -->
 %
 %   Update the Prolog flag `argv`, extracting  the leading script files.
 %   This is called after the C based  parser removed Prolog options such
-%   as ``-q``, ``-f none``, etc. These   options  are availabkle through
+%   as ``-q``, ``-f none``, etc.  These   options  are available through
 %   '$cmd_option_val'/2.
 %
 %   Our task is to update the Prolog flag   `argv`  and return a list of
@@ -556,7 +556,8 @@ hkey('HKEY_LOCAL_MACHINE/Software/SWI/Prolog').
 '$initialise' :-
     catch(initialise_prolog, E, initialise_error(E)).
 
-initialise_error('$aborted') :- !.
+initialise_error(unwind(abort)) :- !.
+initialise_error(unwind(halt(_))) :- !.
 initialise_error(E) :-
     print_message(error, initialization_exception(E)),
     fail.
@@ -567,13 +568,13 @@ initialise_prolog :-
     apple_setup_app,                            % MacOS cwd/locale setup for swipl-win
     init_optimise,
     '$run_initialization',
-    argv_prolog_files(Files, ScriptMode),
     '$load_system_init_file',                   % -F file
     set_toplevel,                               % set `toplevel_goal` flag from -t
     '$set_file_search_paths',                   % handle -p alias=dir[:dir]*
     init_debug_flags,
     start_pldoc,                                % handle --pldoc[=port]
     opt_attach_packs,
+    argv_prolog_files(Files, ScriptMode),
     load_init_file(ScriptMode),                 % -f file
     catch(setup_colors, E, print_message(warning, E)),
     win_associated_files(Files),                % swipl-win: cd and update title
@@ -1047,7 +1048,7 @@ read_expanded_query(BreakLev, ExpandedQuery, ExpandedBindings) :-
 read_query(_Prompt, Goal, Bindings) :-
     '$can_yield',
     !,
-    await(goal, GoalString),
+    await(query, GoalString),
     term_string(Goal, GoalString, [variable_names(Bindings)]).
 :- endif.
 read_query(Prompt, Goal, Bindings) :-
@@ -1397,6 +1398,11 @@ anon_vars([]).
 anon_vars(['$VAR'('_')|T]) :-
     anon_vars(T).
 
+%!  name_var(+Bindings, -Name, +N0, -N) is det.
+%
+%   True when Name is a valid name for   a new variable where the search
+%   is guided by the number N0. Name may not appear in Bindings.
+
 name_var(Bindings, Name, N0, N) :-
     between(N0, infinite, N1),
     I is N1//26,
@@ -1412,7 +1418,7 @@ name_var(Bindings, Name, N0, N) :-
     !,
     N is N1+1.
 
-is_bound([Vars=_|T], Name) :-
+is_bound([binding(Vars,_Value,_Subst)|T], Name) :-
     (   in_vars(Vars, Name)
     ->  true
     ;   is_bound(T, Name)
