@@ -2,9 +2,10 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2004-2020, University of Amsterdam
+    Copyright (c)  2004-2023, University of Amsterdam
                               VU University Amsterdam
 			      CWI, Amsterdam
+			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -84,6 +85,7 @@ Settings
 Address Converstion
 
 	nbio_get_sockaddr()
+	nbio_get_ip()
 	nbio_get_ip4()
 
 Alternative to nbio_read() and nbio_write(), the application program may
@@ -110,6 +112,7 @@ In the Unix version we simply call PL_dispatch() before doing recv() and
 leave the details to this function.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+#define _CRT_SECURE_NO_WARNINGS 1
 #include <config.h>
 
 #if defined(__MINGW32__)
@@ -118,7 +121,7 @@ leave the details to this function.
 #define __finally
 #endif
 
-#if defined(__MINGW32__)
+#if defined(__WINDOWS__)
 #define WINVER 0x0501
 #include <ws2tcpip.h>
 #endif
@@ -153,7 +156,11 @@ leave the details to this function.
 #define closesocket(n) close((n))	/* same on Unix */
 #endif
 
-#ifndef __WINDOWS__
+#ifdef __WINDOWS__
+typedef int os_bufsize_t;
+#define strdup(s) _strdup(s)
+#else
+typedef size_t os_bufsize_t;
 #define INVALID_SOCKET -1
 #endif
 
@@ -167,10 +174,21 @@ leave the details to this function.
 #define SOCKET_ERROR (-1)
 #endif
 
+/* Everybody seems to defined `struct in6_addr` in a different way ...
+ */
+
+#ifndef s6_addr16
+#if defined(s6_words)
+#define s6_addr16 s6_words
+#elif defined(__APPLE__) || defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
+#define s6_addr16 __u6_addr.__u6_addr16
+#endif
+#endif
+
 #define set(s, f)   ((s)->flags |= (f))
 #define clear(s, f) ((s)->flags &= ~(f))
-#define true(s, f)  ((s)->flags & (f))
-#define false(s, f) (!true(s, f))
+#define ison(s, f)  ((s)->flags & (f))
+#define isoff(s, f) (!ison(s, f))
 
 #define PLSOCK_MAGIC  0x38da3f2c
 #define PLSOCK_CMAGIC 0x38da3f2d
@@ -422,10 +440,10 @@ without dispatching if no input is available.
 
 static int
 wait_socket(plsocket *s)
-{ if ( true(s, PLSOCK_DISPATCH) )
+{ if ( ison(s, PLSOCK_DISPATCH) )
   { int fd = s->socket;
 
-    if ( true(s, PLSOCK_NONBLOCK) && !PL_dispatch(fd, PL_DISPATCH_INSTALLED) )
+    if ( ison(s, PLSOCK_NONBLOCK) && !PL_dispatch(fd, PL_DISPATCH_INSTALLED) )
     {
 #ifdef HAVE_POLL
       struct pollfd fds[1];
@@ -496,6 +514,7 @@ nbio_fcntl(nbio_sock_t socket, int op, int arg)
 
 static functor_t FUNCTOR_module2;
 static functor_t FUNCTOR_ip4;
+static functor_t FUNCTOR_ip8;
 static functor_t FUNCTOR_ip1;
 static atom_t ATOM_any;
 static atom_t ATOM_broadcast;
@@ -555,6 +574,7 @@ allocSocket(SOCKET socket)
 
   DEBUG(2, Sdprintf("[%d]: allocSocket(%d) --> %p\n",
 		    PL_thread_self(), socket, p));
+  DEBUG(4, PL_backtrace(10,1));
 
   return p;
 }
@@ -576,9 +596,9 @@ closeSocket(plsocket *s)
 { int rval;
   SOCKET sock;
 
-  DEBUG(2, Sdprintf("Closing %p (%d)\n", s, s->socket));
+  DEBUG(2, Sdprintf("Closing %p (%zd)\n", s, (size_t)s->socket));
   if ( !s || s->magic != PLSOCK_MAGIC )
-  { DEBUG(1, Sdprintf("OOPS: closeSocket(%p) s->magic = %ld\n",
+  { DEBUG(1, Sdprintf("OOPS: closeSocket(%p) s->magic = %d\n",
 		      s, s ? s->magic : 0));
     errno = EINVAL;
     return -1;
@@ -787,7 +807,7 @@ nbio_error(int code, nbio_error_map mapid)
 			 CompoundArg("error", 2),
 			   CompoundArg("socket_error", 2),
 			     AtomArg(symbol),
-			     AtomArg(msg),
+			     AtomMbArg(msg),
 			   PL_VARIABLE) &&
 	   PL_raise_exception(ex)
 	 );
@@ -811,8 +831,9 @@ nbio_init(const char *module)
   initialised = TRUE;
 
   FUNCTOR_module2  = PL_new_functor(PL_new_atom(":"), 2);
-  FUNCTOR_ip4	   = PL_new_functor(PL_new_atom("ip"), 4);
   FUNCTOR_ip1	   = PL_new_functor(PL_new_atom("ip"), 1);
+  FUNCTOR_ip4	   = PL_new_functor(PL_new_atom("ip"), 4);
+  FUNCTOR_ip8	   = PL_new_functor(PL_new_atom("ip"), 8);
   ATOM_any	   = PL_new_atom("any");
   ATOM_broadcast   = PL_new_atom("broadcast");
   ATOM_loopback	   = PL_new_atom("loopback");
@@ -890,7 +911,7 @@ nbio_closesocket(nbio_sock_t socket)
 
   clear(socket, PLSOCK_VIRGIN);
 
-  if ( true(socket, PLSOCK_OUTSTREAM|PLSOCK_INSTREAM) )
+  if ( ison(socket, PLSOCK_OUTSTREAM|PLSOCK_INSTREAM) )
   { int flags = socket->flags;		/* may drop out! */
 
     if ( flags & PLSOCK_INSTREAM )
@@ -1098,34 +1119,64 @@ or the name of a registered port (e.g. 'smtp').
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 int
-nbio_get_sockaddr(term_t Address, struct sockaddr_in *addr, term_t *varport)
+nbio_get_sockaddr(nbio_sock_t socket,
+		  term_t Address, struct sockaddr_storage *storage,
+		  term_t *varport)
 { int port;
+  struct sockaddr_in  *addr4 = (struct sockaddr_in*)storage;
+  struct sockaddr_in6 *addr6 = (struct sockaddr_in6*)storage;
 
-  memset(addr, 0, sizeof(*addr));
-  addr->sin_family = AF_INET;
-  addr->sin_addr.s_addr = INADDR_ANY;
+  switch(socket->domain)
+  { case AF_INET:
+      memset(addr4, 0, sizeof(*addr4));
+      addr4->sin_family = AF_INET;
+      addr4->sin_addr.s_addr = INADDR_ANY;
+      break;
+    case AF_INET6:
+      memset(addr6, 0, sizeof(*addr6));
+      addr6->sin6_family = AF_INET6;
+      break;
+    default:
+      assert(0);
+  }
 
   if ( PL_is_functor(Address, FUNCTOR_module2) )	/* Host:Port */
   { char *hostName;
     term_t arg = PL_new_term_ref();
 
     _PL_get_arg(1, Address, arg);
-    if ( PL_get_atom_chars(arg, &hostName) )
+    if ( PL_get_chars(arg, &hostName, CVT_ATOM|CVT_STRING|CVT_LIST) )
     { struct addrinfo hints;
       struct addrinfo *res;
       int rc;
 
       memset(&hints, 0, sizeof(hints));
-      hints.ai_family = AF_INET;
+      hints.ai_family = socket->domain;
       if ( (rc=getaddrinfo(hostName, NULL, &hints, &res)) !=  0) /* see (*) */
 	return nbio_error(rc, TCP_GAI_ERRNO);
-      assert(res->ai_family == AF_INET);
-      memcpy(&addr->sin_addr,
-	     &((struct sockaddr_in*)res->ai_addr)->sin_addr,
-	     sizeof(addr->sin_addr));
+      switch(socket->domain)
+      { case AF_INET:
+	  if ( res->ai_family != AF_INET )
+	  { freeaddrinfo(res);
+	    return PL_warning("Expected ip4 address");
+	  }
+	  memcpy(&addr4->sin_addr,
+		 &((struct sockaddr_in*)res->ai_addr)->sin_addr,
+		 sizeof(addr4->sin_addr));
+	  break;
+	case AF_INET6:
+	  if ( res->ai_family != AF_INET6 )
+	  { freeaddrinfo(res);
+	    return PL_warning("Expected ip6 address");
+	  }
+	  memcpy(&addr6->sin6_addr,
+		 &((struct sockaddr_in6*)res->ai_addr)->sin6_addr,
+		 sizeof(addr6->sin6_addr));
+	  break;
+      }
       freeaddrinfo(res);
-    } else if ( !nbio_get_ip(arg, &addr->sin_addr) )
-    { return pl_error(NULL, 0, NULL, ERR_ARGTYPE, 1, arg, "atom|ip/4");
+    } else if ( !nbio_get_ip(socket->domain, arg, storage) )
+    { return PL_type_error("host_or_address", arg);
     }
 
     _PL_get_arg(2, Address, arg);
@@ -1138,24 +1189,32 @@ nbio_get_sockaddr(term_t Address, struct sockaddr_in *addr, term_t *varport)
   } else if ( !nbio_get_port(Address, &port) )
     return FALSE;
 
-  addr->sin_port = htons((short)port);
+  switch(socket->domain)
+  { case AF_INET:
+      addr4->sin_port = htons((short)port);
+      break;
+    case AF_INET6:
+      addr6->sin6_port  = htons((short)port);
+      break;
+  }
 
   return TRUE;
 }
 
 
 int
-nbio_get_ip(term_t ip4, struct in_addr *ip)
+nbio_get_ip4(term_t ip4, struct in_addr *ip, int error)
 { uint32_t hip = 0;
 
   if ( PL_is_functor(ip4, FUNCTOR_ip4) )
-  { int i, ia;
+  { int i;
+    unsigned char ua;
     term_t a = PL_new_term_ref();
 
     for(i=1; i<=4; i++)
     { _PL_get_arg(i, ip4, a);
-      if ( PL_get_integer(a, &ia) )
-	hip |= ia << ((4-i)*8);
+      if ( PL_cvt_i_uchar(a, &ua) )
+	hip |= (uint32_t)ua << ((4-i)*8);
       else
 	return FALSE;
     }
@@ -1176,38 +1235,116 @@ nbio_get_ip(term_t ip4, struct in_addr *ip)
       else if ( id == ATOM_loopback )
 	ip->s_addr = INADDR_LOOPBACK;
       else
-	return FALSE;
+	return PL_domain_error("ip4_symbolic_address", a);
 
       return TRUE;
     }
   }
 
-  return FALSE;
+  if ( error )
+    return PL_domain_error("ip4_address", ip4);
+  else
+    return FALSE;
+}
+
+
+static int
+nbio_get_ip6(term_t ip8, struct in6_addr *ip)
+{ if ( PL_is_functor(ip8, FUNCTOR_ip8) )
+  { unsigned short ia;
+    term_t a = PL_new_term_ref();
+
+    for(int i=1; i<=8; i++)
+    { _PL_get_arg(i, ip8, a);
+      if ( PL_cvt_i_ushort(a, &ia) )
+      { ip->s6_addr16[i-1] = htons(ia);
+      } else
+	return FALSE;
+    }
+
+    return TRUE;
+  } else
+    return PL_domain_error("ip6", ip8);
 }
 
 
 int
-nbio_unify_ip4(term_t Ip, unsigned long hip)
+nbio_get_ip(int domain, term_t Ip, struct sockaddr_storage *storage)
+{ switch( domain )
+  { case AF_INET:
+    { struct sockaddr_in *addr = (struct sockaddr_in*)storage;
+      return nbio_get_ip4(Ip, &addr->sin_addr, TRUE);
+    }
+    case AF_INET6:
+    { struct sockaddr_in6 *addr = (struct sockaddr_in6*)storage;
+      return nbio_get_ip6(Ip, &addr->sin6_addr);
+    }
+    case AF_UNSPEC:
+    { struct sockaddr_in  *addr4 = (struct sockaddr_in*)storage;
+      struct sockaddr_in6 *addr6 = (struct sockaddr_in6*)storage;
+
+      return ( nbio_get_ip4(Ip, &addr4->sin_addr, FALSE) ||
+	       nbio_get_ip6(Ip, &addr6->sin6_addr) );
+    }
+    default:
+      assert(0);
+      return FALSE;
+  }
+}
+
+
+
+int
+nbio_unify_ip4(term_t Ip, uint32_t hip)
 { return PL_unify_term(Ip,
 		       PL_FUNCTOR, FUNCTOR_ip4,
-		         IntArg((hip >> 24) & 0xff),
-		         IntArg((hip >> 16) & 0xff),
-		         IntArg((hip >>  8) & 0xff),
-		         IntArg((hip >>  0) & 0xff));
+			 IntArg((hip >> 24) & 0xff),
+			 IntArg((hip >> 16) & 0xff),
+			 IntArg((hip >>  8) & 0xff),
+			 IntArg((hip >>  0) & 0xff));
+}
+
+
+#define NSHORTARG(i) PL_INT, ntohs(i)
+
+static int
+nbio_unify_ip6(term_t t, struct in6_addr *addr)
+{ return PL_unify_term(t,
+		       PL_FUNCTOR, FUNCTOR_ip8,
+			 NSHORTARG(addr->s6_addr16[0]),
+			 NSHORTARG(addr->s6_addr16[1]),
+			 NSHORTARG(addr->s6_addr16[2]),
+			 NSHORTARG(addr->s6_addr16[3]),
+			 NSHORTARG(addr->s6_addr16[4]),
+			 NSHORTARG(addr->s6_addr16[5]),
+			 NSHORTARG(addr->s6_addr16[6]),
+			 NSHORTARG(addr->s6_addr16[7]));
 }
 
 
 int
-nbio_bind(nbio_sock_t socket, struct sockaddr *my_addr, size_t addrlen)
+nbio_unify_addr(term_t t, struct sockaddr *addr)
+{ switch( addr->sa_family )
+  { case AF_INET:
+    { struct sockaddr_in *addr4 = (struct sockaddr_in*)addr;
+      return nbio_unify_ip4(t, ntohl(addr4->sin_addr.s_addr));
+    }
+    case AF_INET6:
+    { struct sockaddr_in6 *addr6 = (struct sockaddr_in6*)addr;
+      return nbio_unify_ip6(t, &addr6->sin6_addr);
+    }
+    default:
+      assert(0);
+      return FALSE;
+  }
+}
+
+int
+nbio_bind(nbio_sock_t socket, struct sockaddr *my_addr, socklen_t addrlen)
 { VALID_SOCKET(socket);
 
-#ifdef __WINDOWS__
-  if ( bind(socket->socket, my_addr, (int)addrlen) )
-#else
   if ( bind(socket->socket, my_addr, addrlen) )
-#endif
-  {
-    nbio_error(GET_ERRNO, TCP_ERRNO);
+  { nbio_error(GET_ERRNO, TCP_ERRNO);
     return -1;
   }
 
@@ -1216,11 +1353,10 @@ nbio_bind(nbio_sock_t socket, struct sockaddr *my_addr, size_t addrlen)
   return 0;
 }
 
-
 int
 nbio_connect(nbio_sock_t socket,
 	     const struct sockaddr *serv_addr,
-	     size_t addrlen)
+	     socklen_t addrlen)
 { VALID_SOCKET(socket);
 
   for(;;)
@@ -1267,12 +1403,17 @@ nbio_accept(nbio_sock_t master, struct sockaddr *addr, socklen_t *addrlen)
       return NULL;
 #endif
 
+    if ( PL_handle_signals() < 0 )
+      return NULL;
+    DEBUG(3, Sdprintf("[%d] calling accept(%d)\n",
+		      PL_thread_self(), master->socket));
     slave = accept(master->socket, addr, addrlen);
+    DEBUG(3, Sdprintf("[%d] accept(%d) --> %d\n",
+		      PL_thread_self(), master->socket, slave));
 
     if ( slave == SOCKET_ERROR )
     { if ( need_retry(GET_ERRNO) )
-      { if ( PL_handle_signals() < 0 )
-	  return NULL;
+      {
 #ifdef __WINDOWS__
         if ( !wait_socket(master) )
           return NULL;
@@ -1289,7 +1430,7 @@ nbio_accept(nbio_sock_t master, struct sockaddr *addr, socklen_t *addrlen)
   s = allocSocket(slave);
   s->flags |= PLSOCK_ACCEPT;
 #ifndef __WINDOWS__
-  if ( true(s, PLSOCK_NONBLOCK) )
+  if ( ison(s, PLSOCK_NONBLOCK) )
     nbio_setopt(s, TCP_NONBLOCK);
 #endif
 
@@ -1331,7 +1472,7 @@ nbio_read(nbio_sock_t socket, char *buf, size_t bufSize)
       return -1;
 #endif
 
-    n = recv(socket->socket, buf, bufSize, 0);
+    n = recv(socket->socket, buf, (os_bufsize_t)bufSize, 0);
 
     if ( n == -1 )
     { if ( need_retry(GET_ERRNO) )
@@ -1366,7 +1507,7 @@ nbio_write(nbio_sock_t socket, char *buf, size_t bufSize)
   while( len > 0 )
   { int n;
 
-    n = send(socket->socket, str, len, 0);
+    n = send(socket->socket, str, (os_bufsize_t)len, 0);
     if ( n < 0 )
     { if ( need_retry(GET_ERRNO) )
       { if ( PL_handle_signals() < 0 )
@@ -1427,13 +1568,13 @@ nbio_close_input(nbio_sock_t socket)
 
   VALID_SOCKET(socket);
 
-  DEBUG(2, Sdprintf("[%d]: nbio_close_input(%d, flags=0x%x)\n",
+  DEBUG(2, Sdprintf("[%d]: nbio_close_input(%p, flags=0x%x)\n",
 		    PL_thread_self(), socket, socket->flags));
-  if ( true(socket, PLSOCK_INSTREAM) )
+  if ( ison(socket, PLSOCK_INSTREAM) )
   { clear(socket, PLSOCK_INSTREAM);
 
     socket->input = NULL;
-    if ( false(socket, (PLSOCK_INSTREAM|PLSOCK_OUTSTREAM)) )
+    if ( isoff(socket, (PLSOCK_INSTREAM|PLSOCK_OUTSTREAM)) )
       rc = closeSocket(socket);
 
     if ( socket->symbol )
@@ -1458,10 +1599,10 @@ nbio_close_output(nbio_sock_t socket)
 
   VALID_SOCKET(socket);
 
-  DEBUG(2, Sdprintf("[%d]: nbio_close_output(%d, flags=0x%x)\n",
+  DEBUG(2, Sdprintf("[%d]: nbio_close_output(%p, flags=0x%x)\n",
 		    PL_thread_self(), socket, socket->flags));
 
-  if ( true(socket, PLSOCK_OUTSTREAM) )
+  if ( ison(socket, PLSOCK_OUTSTREAM) )
   { clear(socket, PLSOCK_OUTSTREAM);
 
     if ( socket->socket != INVALID_SOCKET )
@@ -1471,7 +1612,7 @@ nbio_close_output(nbio_sock_t socket)
     }
 
     socket->output = NULL;
-    if ( false(socket, (PLSOCK_INSTREAM|PLSOCK_OUTSTREAM)) )
+    if ( isoff(socket, (PLSOCK_INSTREAM|PLSOCK_OUTSTREAM)) )
       rc = (rc + closeSocket(socket)) ? -1 : 0;
 
     if ( socket->symbol )
@@ -1500,7 +1641,7 @@ nbio_recvfrom(nbio_sock_t socket, void *buf, size_t bufSize, int flags,
       return -1;
 #endif
 
-    n = recvfrom(socket->socket, buf, bufSize, flags, from, fromlen);
+    n = recvfrom(socket->socket, buf, (os_bufsize_t)bufSize, flags, from, fromlen);
 
     if ( n == -1 )
     { if ( need_retry(GET_ERRNO) )
@@ -1536,7 +1677,7 @@ nbio_sendto(nbio_sock_t socket, void *buf, size_t bufSize, int flags,
   VALID_SOCKET(socket);
 
   for(;;)
-  { n = sendto(socket->socket, buf, bufSize, flags, to, tolen);
+  { n = sendto(socket->socket, buf, (os_bufsize_t)bufSize, flags, to, tolen);
 
     if ( n < 0 )
     { if ( need_retry(GET_ERRNO) )
