@@ -53,6 +53,7 @@
 #include "ParenMatching.h"
 
 #include <QElapsedTimer>
+#include <QShortcut>
 
 class Swipl_IO;
 
@@ -115,7 +116,6 @@ public:
     private:
         QThread *stop_, *go_;
         QMutex sync;
-        int timeout_ms;
     };
 
     /** give access to rl_... predicates */
@@ -247,6 +247,9 @@ protected:
     /** relax selection check requirement */
     QElapsedTimer sel_check_timing;
 
+protected:
+    QShortcut *pasteQuoted = nullptr;
+
 protected:
 
     /** keep last matched pair */
@@ -259,7 +262,7 @@ public slots:
 
     /** serve console menus */
     void onConsoleMenuAction();
-    void onConsoleMenuActionMap(const QString &action);
+    void onConsoleMenuActionMap(QString action);
 
     /** 2. attempt to run generic code inter threads */
     void run_function(pfunc f) { f(); }
