@@ -11,33 +11,43 @@
 
 # Set PREFIX to point at the prefix for installing the dependencies.
 # Configure using cmake -DMACOSX_DEPENDENCIES_FROM=$PREFIX
+#
+# Ideally, we build _universal_  libraries, but this seems complicated
+# because the  configuration of some  of the libraries depends  on the
+# CPU.  E.g., OpenSSL  does not build on the M1  using `-arch x86_64`.
+# pcre   does  not   include  the   JIT  compiler,   etc.   Therefore,
+# unfortunately, we must build the  libraries on a read x86_64 machine
+# and  combine  them  using  the  `macos-import-arch.sh`  script  into
+# universal binaries.
 
 PREFIX="$HOME/deps"
-export MACOSX_DEPLOYMENT_TARGET=10.14
+export MACOSX_DEPLOYMENT_TARGET=10.15
 
-GMP_VERSION=6.2.1
-SSL_VERSION=1.1.1m
-JPEG_VERSION=9b
-ZLIB_VERSION=1.2.11
-ARCHIVE_VERSION=3.5.2
+GMP_VERSION=6.3.0
+SSL_VERSION=3.2.0
+JPEG_VERSION=9f
+ZLIB_VERSION=1.3.1
+ARCHIVE_VERSION=3.7.2
 UUID_VERSION=1.6.2
 BDB_VERSION=6.1.26
-ODBC_VERSION=2.3.9
-PCRE_VERSION=8.45
-PCRE2_VERSION=10.39
-FFI_VERSION=3.4.2
-YAML_VERSION=0.1.7
-READLINE_VERSION=8.0
+ODBC_VERSION=2.3.12
+PCRE2_VERSION=10.42
+FFI_VERSION=3.4.4
+YAML_VERSION=0.2.5
+READLINE_VERSION=8.2
 
 # installation prefix.  This path should not have spaces in one of the
 # directory names.
 
 src="$(pwd)"
 ################
-# Handy for running autoconf from a directory
+# LDFLAGS allows for running autoconf from a directory
 
 export LDFLAGS=-L$PREFIX/lib
-export CFLAGS="-mmacosx-version-min=10.14 -O2"
+export CUFLAGS="-arch x86_64"
+export CMFLAGS="-mmacosx-version-min=$MACOSX_DEPLOYMENT_TARGET -O2"
+export CFLAGS="$CMFLAGS"
+#export CFLAGS="$CUFLAGS $CMFLAGS"
 
 config()
 { if [ -r ./configure ]; then
@@ -88,6 +98,7 @@ build_ssl()
 	   return 1
 	   ;;
     esac
+    export CFLAGS="$CMFLAGS"	# cannot build universal binary
     ./Configure --prefix=$PREFIX shared threads $target
     make depend
     make
@@ -126,7 +137,7 @@ download_odbc()
 { ODBC_FILE=unixODBC-$ODBC_VERSION.tar.gz
 
   [ -f $ODBC_FILE ] || \
-  curl http://www.unixodbc.org/$ODBC_FILE > $ODBC_FILE
+  curl https://www.unixodbc.org/$ODBC_FILE > $ODBC_FILE
   tar zxvf $ODBC_FILE
 }
 
@@ -223,25 +234,6 @@ build_libarchive()
 #################################
 # Download and install libpcre
 
-download_libpcre()
-{ PCRE_FILE=pcre-$PCRE_VERSION.tar.gz
-
-  [ -f $PCRE_FILE ] || \
-    wget https://ftp.pcre.org/pub/pcre/$PCRE_FILE
-  tar xzf $PCRE_FILE
-}
-
-
-build_libpcre()
-{ ( cd pcre-$PCRE_VERSION
-    ./configure --prefix=$PREFIX \
-	--disable-static --disable-cpp --enable-utf8 --enable-unicode-properties
-    make pcre.dll
-    make install
-  )
-}
-
-
 download_libpcre2()
 { PCRE2_FILE=pcre2-$PCRE2_VERSION.tar.gz
 
@@ -332,16 +324,16 @@ build_emacs()
 # Do the whole lot for all prerequisites
 
 clean_prerequisites()
-{ ( cd gmp-$GMP_VERSION && make distclean )
-  ( cd openssl-$SSL_VERSION && make distclean )
-  ( cd jpeg-$JPEG_VERSION && make distclean )
-  ( cd zlib-$ZLIB_VERSION && make distclean )
-  ( cd libarchive-$ARCHIVE_VERSION && make distclean )
-  ( cd uuid-$UUID_VERSION && make distclean )
-  ( cd ffi-$FFI_VERSION && make distclean )
-  ( cd libyaml && make distclean )
-}
+{ rm -rf jpeg-9f
+  for f in *.tar.*; do
+      dir=$(echo $f | sed 's/\.tar\..*//')
+      echo "Cleaning $dir"
+      rm -rf $dir
+      tar zxf $f
+  done
 
+  ( cd libyaml && git clean -xfd )
+}
 
 download_prerequisites()
 { download_gmp
@@ -353,7 +345,7 @@ download_prerequisites()
   download_libuuid
   download_libdb
   download_odbc
-  download_libpcre
+  download_libpcre2
   download_libffi
   download_libyaml
 }
@@ -368,7 +360,7 @@ build_prerequisites()
   build_libuuid
   build_libdb
   build_odbc
-  build_libpcre
+  build_libpcre2
   build_libffi
   build_libyaml
   build_emacs
