@@ -35,7 +35,8 @@
 */
 
 /*#define O_DEBUG 1*/
-#include "pl-incl.h"
+#include "pl-funct.h"
+#include "pl-fli.h"
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Functor (name/arity) handling.  A functor is a unique object (like atoms).
@@ -112,19 +113,17 @@ registerFunctor(FunctorDef fd)
 
   amask = (fd->arity < F_ARITY_MASK ? fd->arity : F_ARITY_MASK);
   fd->functor = MK_FUNCTOR(index, amask);
-  MEMORY_BARRIER();			/* See (*) */
-  fd->flags |= VALID_F;
-  MEMORY_BARRIER();			/* See (*) */
   GD->functors.array.blocks[idx][index] = fd;
+  MEMORY_RELEASE();			/* See (*) */
+  fd->flags |= VALID_F;
 
   DEBUG(CHK_SECURE, assert(fd->arity == arityFunctor(fd->functor)));
 }
 
 
 functor_t
-lookupFunctorDef(atom_t atom, size_t arity)
-{ GET_LD
-  int v;
+lookupFunctorDef(DECL_LD atom_t atom, size_t arity)
+{ int v;
   FunctorDef *table;
   int buckets;
   FunctorDef f, head;
@@ -135,7 +134,7 @@ redo:
   v = (int)pointerHashValue(atom, buckets);
   head = table[v];
 
-  DEBUG(9, Sdprintf("Lookup functor %s/%d = ", stringAtom(atom), arity));
+  DEBUG(9, Sdprintf("Lookup functor %s/%zd = ", stringAtom(atom), arity));
   for(f = table[v]; f; f = f->next)
   { if (atom == f->name && f->arity == arity)
     { DEBUG(9, Sdprintf("%p (old)\n", f));
@@ -201,7 +200,7 @@ static void
 rehashFunctors(void)
 { FunctorTable newtab;
   size_t index;
-  int i, last = FALSE;
+  int i, last = false;
 
   if ( functorDefTable->buckets * 2 >= GD->statistics.functors )
     return;
@@ -216,7 +215,7 @@ rehashFunctors(void)
 	Sdprintf("Rehashing functor-table (%d --> %d)\n",
 		 functorDefTable->buckets, newtab->buckets));
 
-  GD->functors.rehashing = TRUE;
+  GD->functors.rehashing = true;
   for(index=1, i=0; !last; i++)
   { size_t upto = (size_t)2<<i;
     FunctorDef *b;
@@ -226,7 +225,7 @@ rehashFunctors(void)
 
       if ( upto >= high )
       { upto = high;
-	last = TRUE;
+	last = true;
       }
 
       for(; index<upto; index++)
@@ -243,7 +242,7 @@ rehashFunctors(void)
   }
 
   functorDefTable = newtab;
-  GD->functors.rehashing = FALSE;
+  GD->functors.rehashing = false;
   maybe_free_functor_tables();
 }
 
@@ -254,7 +253,10 @@ not known.
 
 functor_t
 isCurrentFunctor(atom_t atom, size_t arity)
-{ GET_LD
+{
+#ifdef O_PLMT
+  GET_LD
+#endif
   unsigned int v;
   int buckets;
   FunctorDef *table;
@@ -461,12 +463,12 @@ checkFunctors()
 }
 #endif
 
-word
+foreign_t
 pl_current_functor(term_t name, term_t arity, control_t h)
 { GET_LD
   atom_t nm = 0;
   size_t index;
-  int i, last=FALSE;
+  int i, last=false;
   int  ar;
   fid_t fid;
 
@@ -474,7 +476,7 @@ pl_current_functor(term_t name, term_t arity, control_t h)
   { case FRG_FIRST_CALL:
       if ( PL_get_atom(name, &nm) &&
 	   PL_get_integer(arity, &ar) )
-	return isCurrentFunctor(nm, ar) ? TRUE : FALSE;
+	return isCurrentFunctor(nm, ar) ? true : false;
 
       if ( !(PL_is_integer(arity) || PL_is_variable(arity)) )
 	return PL_error("current_functor", 2, NULL, ERR_TYPE,
@@ -503,7 +505,7 @@ pl_current_functor(term_t name, term_t arity, control_t h)
 
     if ( upto >= high )
     { upto = high;
-      last = TRUE;
+      last = true;
     }
 
     for(; index<upto; index++)
@@ -523,7 +525,7 @@ pl_current_functor(term_t name, term_t arity, control_t h)
   }
 
   PL_UNLOCK(L_FUNCTOR);
-  return FALSE;
+  return false;
 }
 
 
